#include <oxstd.h>
#include <oxfloat.h>
#include <oxprob.h>

//**********************************************************
// iqrproc(b,B,tau,J,H,b1,b2,y,d,x,inst)
// Performs process tests from Chernozhukov and Hansen (2002).
// Specifically, performs 4 tests.
//		1. Zero treatment effects (Equality of distribution)
//		2. Constant treatment effect (Based on QR coef closest to median)
//		3. Stochastic dominance test
//		4. "Hausman" exogeneity test
//
// Note, all inputs are required.  If there are no covariates or instruments, pass <>.
//  b - subsample size
//  B - number of bootstrap replications
//  tau - vector of quantile indices corresponding to b, B
//  J - Stacked J matrices of IQR where sqrt(n)(b1hat - b1) -> J*N(0,S)
//  H - Stacked H matrices of QR where sqrt(n)(b2hat - b2) -> H*N(0,S)
//	b1 - matrix of IQR estimates (Coeffs and std errors)
//  b2 - matrix of QR estimates (Coeffs and std errors)
//  y - dependent variable
//  d - endogenous RHS variable (Current program requires D be an n X 1 vector)
//  x - covariates
//  inst - excluded instruments
// Output.
//  Printed output is Kolmgorov-Smirnov and Smirnov test statistics and
//     selected critical values.
//  Output is an array with the test statistics and values of the bootstrap
//     replications.  The output is arranged with the first four elements of the
//	   array corresponding to the zero effect hypothesis, the next four for the
//     constant effect hypothesis, etc.  Within hypothesis, the elements are
//     arranged as CM test stat, KS test stat, CM bootstrap values, KS bootstrap
//     values.
//*****************************************************************************

iqrproc(b,B,tau,J,H,b1,b2,y,d,x,inst){

	decl i, sampid, zb, vb, result, z, omega, temp1, vn, u, v, j,
		 CM, CMb, K, Kb, CM90, CM95, CM99, K90, K95, K99;

	b = floor(b);
	x = x~ones(rows(y),1);
	decl index = range(0,columns(b1)-2,2);
	decl beta1 = (b1[0:columns(d)-1][index])';
	decl beta2 = (b2[0:columns(d)-1][index])';

	decl R = ones(1,1)~zeros(1,columns(J)-1);
	decl t = columns(tau);
	decl n = rows(y);

//  Test for equality of distribution

	z = zeros(1,n);
	for (i = 0; i < t; ++i)
	{
	z = z | R*(((J[(i)*(columns(J)):(i+1)*columns(J)-1][])')*
		((tau[][i]-((y-(d~x)*b1[][2*i]).<0))'.*(x~inst)'));
	}
	z = z[1:][];
	vn = meanr(z);
	omega = meanr(z.^2);
	CM = n*meanc(beta1.*((omega.^(-1)).*beta1));
	K = sqrt(n)*maxc(sqrt(beta1.*((omega.^(-1)).*beta1)));
	CMb = zeros(1,B);
	Kb = zeros(1,B);
	for (i = 0; i < B; ++i)
	{
		sampid = floor((ranu(b,1))*n);
		zb = z[][sampid];
		vb = meanr(zb);
		CMb[][i] = b*meanc(vb.*((omega.^(-1)).*vb));
		Kb[][i] = sqrt(b)*maxc(sqrt(vb.*((omega.^(-1)).*vb)));
	}

	CM90 = quantiler(CMb,.90);
	CM95 = quantiler(CMb,.95);
	CM99 = quantiler(CMb,.99);
	K90 = quantiler(Kb,.90);
	K95 = quantiler(Kb,.95);
	K99 = quantiler(Kb,.99);

	println("Equality of distribution test:\n");
	println("CM Test");
	println("Test statistic:  ",CM,"\n");
	println("Critical Values:\n");
	println("    90%:  ",CM90,"\n");
	println("    95%:  ",CM95,"\n");
	println("    99%:  ",CM99,"\n");

	println("KS Test");
	println("Test statistic:  ",K,"\n");
	println("Critical Values:\n");
	println("    90%:  ",K90,"\n");
	println("    95%:  ",K95,"\n");
	println("    99%:  ",K99,"\n");

	result = new array[16];
	result[0] = CM;
	result[1] = K;
	result[2] = CMb;
	result[3] = Kb;


//  Test for constant treatment effect based on median (or closest to median)
//  regression result.

	decl in = (fabs(tau - .50001) .== min(fabs(tau - .50001)));
	decl bc = selectifc(b1[][index],in);
	decl tc = selectifc(tau,in);
	decl temp = vecindex(in);
	decl Jc = J[(temp)*(columns(J)):(temp+1)*columns(J)-1][];

	z = zeros(1,n);
	for (i = 0; i < t; ++i)
	{
	z = z | R*(((J[(i)*(columns(J)):(i+1)*columns(J)-1][])')*
		((tau[][i]-((y-(d~x)*b1[][2*i]).<0))'.*(x~inst)')
		 - (Jc'*((tc - ((y-(d~x)*bc) .< 0))'.*(x~inst)')));
	}
	z = z[1:][];
	vn = meanr(z);
	omega = meanr(z.^2);
	omega = deleteifr(omega,in');
	temp = deleteifr(beta1,in');
	CM = n*meanc(((temp-bc[0][0]).*((omega.^(-1)).*(temp-bc[0][0]))) | 0);
	K = sqrt(n)*maxc(sqrt(((temp-bc[0][0]).*((omega.^(-1)).*(temp-bc[0][0]))) | 0));
	CMb = zeros(1,B);
	Kb = zeros(1,B);
	for (i = 0; i < B; ++i)
	{
		sampid = floor((ranu(b,1))*n);
		zb = z[][sampid];
		vb = meanr(zb);
		vb = deleteifr(vb,in');
		CMb[][i] = b*meanc((vb.*((omega.^(-1)).*vb)) | 0);
		Kb[][i] = sqrt(b)*maxc(sqrt((vb.*((omega.^(-1)).*vb)) | 0));
	}

	CM90 = quantiler(CMb,.90);
	CM95 = quantiler(CMb,.95);
	CM99 = quantiler(CMb,.99);
	K90 = quantiler(Kb,.90);
	K95 = quantiler(Kb,.95);
	K99 = quantiler(Kb,.99);

	println("Constant effect test (based on median):\n");
	println("CM Test");
	println("Test statistic:  ",CM,"\n");
	println("Critical Values:\n");
	println("    90%:  ",CM90,"\n");
	println("    95%:  ",CM95,"\n");
	println("    99%:  ",CM99,"\n");

	println("KS Test");
	println("Test statistic:  ",K,"\n");
	println("Critical Values:\n");
	println("    90%:  ",K90,"\n");
	println("    95%:  ",K95,"\n");
	println("    99%:  ",K99,"\n");
	
	result[4] = CM;
	result[5] = K;
	result[6] = CMb;
	result[7] = Kb;


//  Test for stochastic dominance

	z = zeros(1,n);
	for (i = 0; i < t; ++i)
	{
	z = z | R*(((J[(i)*(columns(J)):(i+1)*columns(J)-1][])')*
		((tau[][i]-((y-(d~x)*b1[][2*i]).<0))'.*(x~inst)'));
	}
	z = z[1:][];
	vn = meanr(z);
	omega = meanr(z.^2);
	temp = meanc(beta1);
	if(temp >= 0){
		temp1 = maxc((-beta1~zeros(rows(beta1),1))');
		temp1 = temp1';
		CM = n*meanc(temp1.*((omega.^(-1)).*temp1));
		K = sqrt(n)*maxc((omega.^(-1/2)).*temp1);
		CMb = zeros(1,B);
		Kb = zeros(1,B);
		for (i = 0; i < B; ++i)
		{
			sampid = floor((ranu(b,1))*n);
			zb = z[][sampid];
			vb = meanr(zb);
			temp1 = maxc((-vb~zeros(rows(vb),1))');
			temp1 = temp1';
			CMb[][i] = b*meanc(temp1.*((omega.^(-1)).*temp1));
			Kb[][i] = sqrt(b)*maxc((omega.^(-1/2)).*temp1);
		}
	}
	else{
		temp1 = maxc((beta1~zeros(rows(beta1),1))');
		temp1 = temp1';
		CM = n*meanc(temp1.*((omega.^(-1)).*temp1));
		K = sqrt(n)*maxc((omega.^(-1/2)).*temp1);
		CMb = zeros(1,B);
		Kb = zeros(1,B);
		for (i = 0; i < B; ++i)
		{
			sampid = floor((ranu(b,1))*n);
			zb = z[][sampid];
			vb = meanr(zb);
			temp1 = maxc((vb~zeros(rows(vb),1))');
			temp1 = temp1';
			CMb[][i] = b*meanc(temp1.*((omega.^(-1)).*temp1));
			Kb[][i] = sqrt(b)*maxc((omega.^(-1/2)).*temp1);
		}
	}		

	CM90 = quantiler(CMb,.90);
	CM95 = quantiler(CMb,.95);
	CM99 = quantiler(CMb,.99);
	K90 = quantiler(Kb,.90);
	K95 = quantiler(Kb,.95);
	K99 = quantiler(Kb,.99);

	println("Stochastic dominance:\n");
	println("CM Test");
	println("Test statistic:  ",CM,"\n");
	println("Critical Values:\n");
	println("    90%:  ",CM90,"\n");
	println("    95%:  ",CM95,"\n");
	println("    99%:  ",CM99,"\n");

	println("KS Test");
	println("Test statistic:  ",K,"\n");
	println("Critical Values:\n");
	println("    90%:  ",K90,"\n");
	println("    95%:  ",K95,"\n");
	println("    99%:  ",K99,"\n");
	
	result[8] = CM;
	result[9] = K;
	result[10] = CMb;
	result[11] = Kb;

//  Test for exogeneity

	z = zeros(1,n);
	for (i = 0; i < t; ++i)
	{
	z = z | R*(((J[(i)*(columns(J)):(i+1)*columns(J)-1][])')*
		((tau[][i]-((y-(d~x)*b1[][2*i]).<0))'.*(x~inst)')
		 - ((H[(i)*(columns(H)):(i+1)*columns(H)-1][])')*
		((tau[][i]-((y-(d~x)*b2[][2*i]).<0))'.*(d~x)'));
	}
	z = z[1:][];
	vn = meanr(z);
	omega = meanr(z.^2);
	temp = beta1-beta2;
	CM = n*meanc(temp.*((omega.^(-1)).*temp));
	K = sqrt(n)*maxc(sqrt(temp.*((omega.^(-1)).*temp)));
	CMb = zeros(1,B);
	Kb = zeros(1,B);
	for (i = 0; i < B; ++i)
	{
		sampid = floor((ranu(b,1))*n);
		zb = z[][sampid];
		vb = meanr(zb);
		CMb[][i] = b*meanc(vb.*((omega.^(-1)).*vb));
		Kb[][i] = sqrt(b)*maxc(sqrt(vb.*((omega.^(-1)).*vb)));
	}

	CM90 = quantiler(CMb,.90);
	CM95 = quantiler(CMb,.95);
	CM99 = quantiler(CMb,.99);
	K90 = quantiler(Kb,.90);
	K95 = quantiler(Kb,.95);
	K99 = quantiler(Kb,.99);

	println("Exogeneity:\n");
	println("CM Test");
	println("Test statistic:  ",CM,"\n");
	println("Critical Values:\n");
	println("    90%:  ",CM90,"\n");
	println("    95%:  ",CM95,"\n");
	println("    99%:  ",CM99,"\n");

	println("KS Test");
	println("Test statistic:  ",K,"\n");
	println("Critical Values:\n");
	println("    90%:  ",K90,"\n");
	println("    95%:  ",K95,"\n");
	println("    99%:  ",K99,"\n");
	
	result[12] = CM;
	result[13] = K;
	result[14] = CMb;
	result[15] = Kb;

	return(result);

}