#include <oxstd.h>
#include <oxfloat.h>
#include <oxprob.h>

//******************************************************************************
// Function: Kmat(e,x,z,kernel,h,A,tau,alpha)
//
// Estimates K (where V(b) = inv(J'inv(K)*J)J'inv(K)*S*inv(K)*J*inv(J'inv(K)*J))
//   using kernel methods a la Powell (1984) for
//   use in the variance matrix of IQR estimates.

Kmat(e,x,z,kernel,h,A,tau,alpha){

	decl res, n, psi, J_t, invJ_t, F22, Jbb, Jxz, Kx, Kz, g, k, x0;

	res = new array[3];
	
	n = rows(e);
	k = columns(x);
	
	if(A == 0){
	    A = (z'*z) - (z'*x)*(invert(x'*x))*(x'*z);
	}

	psi = x~z;

	if(kernel == 1){ 		// Gaussian kernel
    
    	if(h == 0){
        	h = (quantilec(e,.75)-quantilec(e,.25))*(n^(-2/5));
		                      // Plug-in method used in Morillo's ox code
                              // for Koenker & Portnoy (1997)
		}
    	else if(h == -1){
        	h = 1.364*((2*sqrt(M_PI))^(-1/5))*sqrt(varc(e))*(n^(-1/5));
                              // Silverman's rule of thumb -- Silverman (1986)
                              // in Hardle & Linton (1994)
    	}
		else if(h == -2){
			x0 = quann(tau);
			h = (n^(-1/3))*(quann(1 - alpha/2)^(2/3))*((1.5*(densn(x0))^2)/(2*x0^2+1))^(1/3);
			                  // Hall and Sheather bandwidth (1988)
		}
		else if(h == -3){
			x0 = quann(tau);
			h = (n^-0.2)*((4.5*(densn(x0))^4)/(2*x0^2+1)^2)^0.2;
			                  // Bofinger (1975)
		}

        J_t = (1/h)*(densn(e/h).*psi)'*psi;

		invJ_t = invert(J_t);

		while(invJ_t == 0)
		{
			h = 2*h;
	        J_t = (1/h)*(densn(e/h).*psi)'*psi;
			invJ_t = invert(J_t);
		}
    
    	if(A == 1){
        	g = (1/n)*J_t;
		}
    	else{
        	F22 = invJ_t[k:rows(J_t)-1][k:rows(J_t)-1];
        	Jbb = J_t[0:k-1][0:k-1];
        	Jxz = J_t[0:k-1][k:rows(J_t)-1];
        	Kx = Jbb ~ Jxz;
        	Kz = Jxz' ~ (invert(F22*A*F22) + Jxz'*invert(Jbb)*Jxz);
        	g = (1/n)*(Kx | Kz);
    	}
    }    
	else if(kernel == 0){ 	// Uniform kernel
    
    	if(h == 0){
        	h = 1.740*(quantilec(e,.75)-quantilec(e,.25))*(n^(-2/5));
							  // Plug-in from Morillo adjusted for 
                              // uniform kernel
		}
    	else if(h == -1){
        	h = 1.364*((9/2)^(1/5))*sqrt(varc(e))*(n^(-1/5));
                              // Silverman's rule of thumb    
    	}
		else if(h == -2){
			x0 = quann(tau);
			h = (n^(-1/3))*(quann(1 - alpha/2)^(2/3))*((1.5*(densn(x0))^2)/(2*x0^2+1))^(1/3);
			                  // Hall and Sheather bandwidth (1988)
		}
		else if(h == -3){
			x0 = quann(tau);
			h = (n^-0.2)*((4.5*(densn(x0))^4)/(2*x0^2+1)^2)^0.2;
			                  // Bofinger (1975)
		}

		J_t = (1/h)*((1/2)*(fabs(e)<=h).*psi)'*psi;

		invJ_t = invert(J_t);

		while(invJ_t == 0)
		{
			h = 2*h;
	        J_t = (1/h)*(densn(e/h).*psi)'*psi;
			invJ_t = invert(J_t);
		}

		if(A == 1){
        	g = (1/n)*J_t;
		}
    	else{
        	invJ_t = invert(J_t);
        	F22 = invJ_t[k:rows(J_t)-1][k:rows(J_t)-1];
        	Jbb = J_t[0:k-1][0:k-1];
        	Jxz = J_t[0:k-1][k:rows(J_t)-1];
        	Kx = Jbb ~ Jxz;
        	Kz = Jxz' ~ (invert(F22*A*F22) + Jxz'*invert(Jbb)*Jxz);
        	g = (1/n)*(Kx | Kz);
    	}
    }    

	res[0] = g;
	res[1] = J_t;
	res[2] = h;
	
	return(res);

}



//******************************************************************************
// Function: Jmat(e,w,psi,kernel,h,tau,alpha)
//
// Estimates E[f(0|X,Z,D)W*PSI'] using kernel methods a la Powell (1984) for
//   use in the variance matrix of IQR estimates.

Jmat(e,w,psi,kernel,h,tau,alpha){

	decl f, n, x0;

	n = rows(e);

	if(kernel == 1){          // Gaussian kernel
    
    	if(h == 0){
        	h = (quantilec(e,.75)-quantilec(e,.25))*(n^(-2/5));
		                      // Plug-in method used in Morillo's ox code
                              // for Koenker & Portnoy (1997)
		}
    	else if(h == -1){
        	h = 1.364*((2*sqrt(M_PI))^(-1/5))*sqrt(varc(e))*(n^(-1/5));
                              // Silverman's rule of thumb -- Silverman (1986)
                              // in Hardle & Linton (1994)
    	}
		else if(h == -2){
			x0 = quann(tau);
			h = (n^(-1/3))*(quann(1 - alpha/2)^(2/3))*((1.5*(densn(x0))^2)/(2*x0^2+1))^(1/3);
			                  // Hall and Sheather bandwidth (1988)
		}
		else if(h == -3){
			x0 = quann(tau);
			h = (n^-0.2)*((4.5*(densn(x0))^4)/(2*x0^2+1)^2)^0.2;
			                  // Bofinger (1975)
		}
    
    	f = (1/(n*h))*(densn(e/h).*w)'*psi;
    }
	else if(kernel == 0){     // Uniform kernel
    
    	if(h == 0){
        	h = 1.740*(quantilec(e,.75)-quantilec(e,.25))*(n^(-2/5));
							  // Plug-in from Morillo adjusted for 
                              // uniform kernel
		}
    	else if(h == -1){
        	h = 1.364*((9/2)^(1/5))*sqrt(varc(e))*(n^(-1/5));
                              // Silverman's rule of thumb    
    	}
		else if(h == -2){
			x0 = quann(tau);
			h = (n^(-1/3))*(quann(1 - alpha/2)^(2/3))*((1.5*(densn(x0))^2)/(2*x0^2+1))^(1/3);
			                  // Hall and Sheather bandwidth (1988)
		}
		else if(h == -3){
			x0 = quann(tau);
			h = (n^-0.2)*((4.5*(densn(x0))^4)/(2*x0^2+1)^2)^0.2;
			                  // Bofinger (1975)
		}
        
    	f = (1/(n*h))*((1/2)*(fabs(e)<=h).*w)'*psi;
    }

	return(f);

}




//******************************************************************************
// Function: kcov_iqr
//
// [b,vc,g,vc_g] = kcov_iqr(beta,gamma,tau,y,d,x,z,kernel,h,A)
// Calculates the heteroskedasticity consistent variance-covariance
//   matrix for inverse quantile regression estimates using kernel methods
//   as introduced in Powell (1984) and Buchinsky (1995).
// Output --
//   b - coefficient vector and standard errors
//       If betas for more than one quantile are passed, the odd rows
//       are the coefficient estimates and the even rows are the standard
//       errors.
//   vc - the complete variance-covariance matrix
//       If more than one quantile is passed this is the variance-covariance
//       matrix of the joint distribution for all of the quantiles passed.
//   g - gamma vector with standard errors.  Output formatted as b.
//   NOTE:  To skip calculation of standard errors for gamma (or if equation
//          is just identified) pass gamma = < >.
//   vc_g - as vc but for gammas
//   NOTE:  Covariance between betas and gammas is not reported.
//
// Input --
//   beta - coefficient estimates from IQR.  Estimates from more than one
//          quantile may be passed.  Each column should contain the estimates
//          for one quantile.
//   gamma - coefficients on instruments in IQR.  Estimates from more than
//          one quantile may be passed.  Each column should contain the estimates
//          for one quantile.  If the system is just identified or standard
//          errors for gamma are not required, pass gamma = < >;
//   tau - row vector containing the quantile index of each column of beta/gamma
//   y - observations on dependent variable (n X 1)
//   d - observations on endogenous variables (n X s);
//   x - observations on covariates (n X k)
//   NOTE:  x should not include a constant term.
//   z - observations on excluded instruments (n X r);
//   kernel - 0 -- uniform kernel
//          - 1 -- gaussian kernel
//   h - bandwidth
//          - 0 -- Plug-in method based on interquartile range
//          - -1 -- Silverman's rule of thumb
//          - -2 -- Hall and Sheather bandwidth
//          - -3 -- Bofinger bandwidth
//   A - weighting matrix
//          - 1 -- Estimates were formed using optimal weighting matrix.  Use
//                   this if equation is just identified.
//          - 0 -- Estimates were formed using optimal weighting matrix under
//                  iid error structure

kcov_iqr(beta,gamma,tau,y,d,x,z,kernel,h,A){

	decl b, vc, g, vc_g, n, i, j, e1, e2, J1, J2, K1, JT1, K2, JT2,
	     S, G1, G2, H, IJT1, C1, IJT2, C2, e, J, K, JT, G, IJT, C,
		 J_t, V, k, t, r, s, result, hnew, k1;

	result = new array[5];

	n = rows(y);	 
	x = x~ones(n,1);
	k = rows(beta);
	t = max(1,columns(tau));
	r = columns(z);
	s = columns(d);
	k1 = columns(x);

    vc = zeros(k*t,k*t);
	b = zeros(k,2*columns(beta));
	vc_g = zeros(r*t,r*t);
	g = zeros(r,2*columns(gamma));
    S = (1/n)*(x~z)'*(x~z);

	for (i = 0; i<t; ++i)
	{
    	for (j = i; j<t; ++j)
		{
        if(i != j){
            e1 = y - (d~x)*(beta[][i]);
            e2 = y - (d~x)*(beta[][j]);
            V = Kmat(e1,x,z,kernel,h,A,tau[0][i],.05);
			K1 = V[0];
			JT1 = V[1];
			hnew = V[2];
			J1 = Jmat(e1,d~x,x~z,kernel,hnew,tau[0][i],.05);
            V = Kmat(e2,x,z,kernel,h,A,tau[0][j],.05);
			K2 = V[0];
			JT2 = V[1];
			hnew = V[2];
            J2 = Jmat(e2,d~x,x~z,kernel,hnew,tau[0][j],.05);
            G1 = invert(J1*invert(K1)*J1');
            G2 = invert(J2*invert(K2)*J2');
            H = J1*invert(K1)*S*invert(K2)*J2';
            vc[k*i:k*(i+1)-1][k*j:k*(j+1)-1] = 
               (1/n)*(min(tau[0][i],tau[0][j])-(tau[0][i])*(tau[0][j]))*G1*H*G2;
            vc[k*j:k*(j+1)-1][k*i:k*(i+1)-1] = 
               (1/n)*(min(tau[0][i],tau[0][j])-(tau[0][i])*(tau[0][j]))*G2'*H'*G1';
            if(rows(gamma) == 0){
                g = < >;
                vc_g = < >;
			}
            else{
                IJT1 = invert(JT1);
				J = IJT1[k1:columns(JT1)-1][];
				G = J[0:s-1][];
				H = invert(K1)-((invert(JT1[0:k1-1][0:k1-1]) ~ zeros(k1,r)) |
						 (zeros(r,k1) ~ zeros(r,r)));
				C1 = J*(unit(k1+r) - G'*invert(G*H*G')*G*H);
                IJT2 = invert(JT2);
				J = IJT2[k1:columns(JT2)-1][];
				G = J[0:s-1][];
				H = invert(K2)-((invert(JT2[0:k1-1][0:k1-1]) ~ zeros(k1,r)) |
						 (zeros(r,k1) ~ zeros(r,r)));
				C2 = J*(unit(k1+r) - G'*invert(G*H*G')*G*H);
                vc_g[r*i:r*(i+1)-1][r*j:r*(j+1)-1] = 
                     (1/n)*(min(tau[0][i],tau[0][j])-tau[0][i]*tau[0][j])*C1*S*C2';
                vc_g[r*j:r*(j+1)-1][r*i:r*(i+1)-1] = 
                     (1/n)*(min(tau[0][i],tau[0][j])-tau[0][i]*tau[0][j])*C2*S*C1';
            }
		}	
        else{
            e = y - (d~x)*beta[][i];
			println(i);
            V = Kmat(e,x,z,kernel,h,A,tau[0][i],.05);
			K = V[0];
			JT = V[1];
			hnew = V[2];
			J = Jmat(e,d~x,x~z,kernel,hnew,tau[0][i],.05);
            G = invert(J*invert(K)*J');
            H = J*invert(K)*S*invert(K)*J';
            vc[k*i:k*(i+1)-1][k*i:k*(i+1)-1] = 
               (1/n)*(tau[0][i]-(tau[0][i])^2)*G*H*G;
            b[][2*i] = beta[][i];
            b[][2*i+1] = (sqrt(diagonal(vc[k*i:k*(i+1)-1][k*i:k*(i+1)-1])))';
            if(rows(gamma) == 0){
                g = < >;
                vc_g = < >;
			}
            else{
                IJT = invert(JT);
				J = IJT[k1:columns(JT)-1][];
				G = J[0:s-1][];
				H = invert(K)-((invert(JT[0:k1-1][0:k1-1]) ~ zeros(k1,r)) |
						 (zeros(r,k1) ~ zeros(r,r)));
				C = J*(unit(k1+r) - G'*invert(G*H*G')*G*H);
                vc_g[r*i:r*(i+1)-1][r*i:r*(i+1)-1] =
                     (1/n)*(tau[0][i]-(tau[0][i])^2)*C*S*C';
                g[][2*i] = gamma[][i];
                g[][2*i+1] = (sqrt(diagonal(vc_g[r*i:r*(i+1)-1][r*i:r*(i+1)-1])))';
            }
			if(i == 0){
				J_t = JT;
			}
			else{
				J_t = J_t | JT;
			}
        }      
    	}
	}

	result[0] = b;
	result[1] = vc;
	result[2] = g;
	result[3] = vc_g;
	result[4] = J_t;

	return(result);

}