#include <oxstd.h>
#include <oxfloat.h>
#include <oxprob.h>


//******************************************************************************
// Function: Jmat(e,x,kernel,h)
//
// Estimates E[f(0|X,Z,D)x*x'] using kernel methods a la Powell (1984) for
//   use in the variance matrix of IQR estimates.

Jmat(e,x,kernel,h){

	decl f, n;

	n = rows(e);

	if(kernel == 1){          // Gaussian kernel
    
    	if(h == 0){
        	h = (quantilec(e,.75)-quantilec(e,.25))*(n^(-2/5));
		                      // Plug-in method used in Morillo's ox code
                              // for Koenker & Portnoy (1997)
		}
    	else if(h < 0){
        	h = 1.364*((2*sqrt(M_PI))^(-1/5))*sqrt(varc(e))*(n^(-1/5));
                              // Silverman's rule of thumb -- Silverman (1986)
                              // in Hardle & Linton (1994)
    	}
    
    	f = ((1/(n*h))*(densn(e/h).*x)'*x)^(-1);
    }
	else if(kernel == 0){     // Uniform kernel
    
    	if(h == 0){
        	h = 1.740*(quantilec(e,.75)-quantilec(e,.25))*(n^(-2/5));
							  // Plug-in from Morillo adjusted for 
                              // uniform kernel
		}
    	else if(h < 0){
        	h = 1.364*((9/2)^(1/5))*sqrt(varc(e))*(n^(-1/5));
                              // Silverman's rule of thumb    
    	}
        
    	f = ((1/(n*h))*((1/2)*(fabs(e)<=h).*x)'*x)^(-1);
    }

	return(f);

}




//******************************************************************************
// Function: kcov_qr
//
// [b,vc] = kcov_iqr(beta,tau,y,x,kernel,h,A)
// Calculates the heteroskedasticity consistent variance-covariance
//   matrix for inverse quantile regression estimates using kernel methods
//   as introduced in Powell (1984) and Buchinsky (1995).
// Output --
//   b - coefficient vector and standard errors
//       If betas for more than one quantile are passed, the odd rows
//       are the coefficient estimates and the even rows are the standard
//       errors.
//   vc - the complete variance-covariance matrix
//       If more than one quantile is passed this is the variance-covariance
//       matrix of the joint distribution for all of the quantiles passed.
//
// Input --
//   beta - coefficient estimates from IQR.  Estimates from more than one
//          quantile may be passed.  Each column should contain the estimates
//          for one quantile.
//   tau - row vector containing the quantile index of each column of beta/gamma
//   y - observations on dependent variable (n X 1)
//   x - observations on covariates (n X k)
//   NOTE:  x should not include a constant term.
//   kernel - 0 -- uniform kernel
//          - 1 -- gaussian kernel
//   h - bandwidth
//          - 0 -- Plug-in method based on interquartile range
//          - -1 -- Silverman's rule of thumb
//   A - weighting matrix
//          - 1 -- Estimates were formed using optimal weighting matrix.
//          - 0 -- Estimates were formed using optimal weighting matrix under
//                  iid error structure

kcov_qr(beta,tau,y,x,kernel,h){

	decl b, vc, n, i, j, e1, e2, J1, J2,
	     S, e, J, k, t, result;

	result = new array[2];

	n = rows(y);	 
	x = x~ones(n,1);
	k = rows(beta);
	t = max(1,columns(tau));

    vc = zeros(k*t,k*t);
	b = zeros(k,2*columns(beta));
    S = (1/n)*(x)'*(x);

	for (i = 0; i<t; ++i)
	{
    	for (j = i; j<t; ++j)
		{
        if(i != j){
            e1 = y - (x)*(beta[][i]);
            e2 = y - (x)*(beta[][j]);
			J1 = Jmat(e1,x,kernel,h);
            J2 = Jmat(e2,x,kernel,h);
            vc[k*i:k*(i+1)-1][k*j:k*(j+1)-1] = 
               (1/n)*(min(tau[0][i],tau[0][j])-(tau[0][i])*(tau[0][j]))*J1*S*J2;
            vc[k*j:k*(j+1)-1][k*i:k*(i+1)-1] = 
               (1/n)*(min(tau[0][i],tau[0][j])-(tau[0][i])*(tau[0][j]))*J2*S*J1;
		}	
        else{
            e = y - (x)*beta[][i];
			println(i);
			J = Jmat(e,x,kernel,h);
            vc[k*i:k*(i+1)-1][k*i:k*(i+1)-1] = 
               (1/n)*(tau[0][i]-(tau[0][i])^2)*J*S*J;
            b[][2*i] = beta[][i];
            b[][2*i+1] = (sqrt(diagonal(vc[k*i:k*(i+1)-1][k*i:k*(i+1)-1])))';
        }      
    	}
	}

	result[0] = b;
	result[1] = vc;

	return(result);

}