#include <oxstd.h>
#include <oxfloat.h>
#include <oxprob.h>

//******************************************************************************
// Function: Kmat(e,x,z,A,tau)
//
// Estimates K (where V(b) = inv(J'inv(K)*J)J'inv(K)*S*inv(K)*J*inv(J'inv(K)*J))
//   using a normal approximation for
//   use in the variance matrix of IQR estimates.

Kmat(e,x,z,A,tau){

	decl res, n, psi, J_t, invJ_t, F22, Jbb, Jxz, Kx, Kz, g, k;

	res = new array[2];
	
	n = rows(e);
	k = columns(x);
	
	if(A == 0){
	    A = (1/n)*((z'*z) - (z'*x)*(invert(x'*x))*(x'*z));
	}

	psi = x~z;

	J_t = (1/n)*(e.*psi)'*psi;
	invJ_t = invert(J_t);
    
   	if(A == 1){
       	g = J_t;
	}
   	else{
       	F22 = invJ_t[k:rows(J_t)-1][k:rows(J_t)-1];
       	Jbb = J_t[0:k-1][0:k-1];
       	Jxz = J_t[0:k-1][k:rows(J_t)-1];
       	Kx = Jbb ~ Jxz;
       	Kz = Jxz' ~ (invert(F22*A*F22) + Jxz'*invert(Jbb)*Jxz);
       	g = Kx | Kz;
   	}

	res[0] = g;
	res[1] = J_t;
	
	return(res);

}



//******************************************************************************
// Function: Jmat(e,w,psi,tau)
//
// Estimates E[f(0|X,Z,D)W*PSI'] using a normal approximation for
//   use in the variance matrix of IQR estimates.

Jmat(e,w,psi,tau){

	decl f, n;

	n = rows(e);

   	f = (1/n)*(e.*w)'*psi;

	return(f);

}



//******************************************************************************
// Function: ncov_iqr
//
// [b,vc,g,vc_g] = ncov_iqr(beta,gamma,tau,y,d,x,z,B50,B75_B25,A)
// Calculates the variance-covariance
//   matrix for inverse quantile regression estimates
//   using a normal approximation.
// Output --
//   b - coefficient vector and standard errors
//       If betas for more than one quantile are passed, the odd rows
//       are the coefficient estimates and the even rows are the standard
//       errors.
//   vc - the complete variance-covariance matrix
//       If more than one quantile is passed this is the variance-covariance
//       matrix of the joint distribution for all of the quantiles passed.
//   g - gamma vector with standard errors.  Output formatted as b.
//   NOTE:  To skip calculation of standard errors for gamma (or if equation
//          is just identified) pass gamma = < >.
//   vc_g - as vc but for gammas
//   NOTE:  Covariance between betas and gammas is not reported.
//
// Input --
//   beta - coefficient estimates from IQR.  Estimates from more than one
//          quantile may be passed.  Each column should contain the estimates
//          for one quantile.
//   gamma - coefficients on instruments in IQR.  Estimates from more than
//          one quantile may be passed.  Each column should contain the estimates
//          for one quantile.  If the system is just identified or standard
//          errors for gamma are not required, pass gamma = < >;
//   tau - row vector containing the quantile index of each column of beta/gamma
//   y - observations on dependent variable (n X 1)
//   d - observations on endogenous variables (n X s);
//   x - observations on covariates (n X k)
//   NOTE:  x should not include a constant term.
//   z - observations on excluded instruments (n X r);
//   B50 - estimated coefficients at median
//   B75_b25 - beta(.75)-beta(.25)
//   A - weighting matrix
//          - 1 -- Estimates were formed using optimal weighting matrix.  Use
//                   this if equation is just identified.
//          - 0 -- Estimates were formed using optimal weighting matrix under
//                  iid error structure

ncov_iqr(beta,gamma,tau,y,d,x,z,B50,B75_B25,A){

	decl b, vc, g, vc_g, n, i, j, yhat1, yhat2, J1, J2, K1, JT1, K2, JT2,
	     S, G1, G2, H, IJT1, C1, IJT2, C2, yhat, J, K, JT, G, IJT, C,
		 J_t, V, k, t, r, s, result, k1, MED, IQRANGE, uhat, ubar, varu, sparsity;

	result = new array[5];
	
	n = rows(y);	 
	x = x~ones(n,1);
	k = rows(beta);
	t = max(1,columns(tau));
	r = columns(z);
	s = columns(d);
	k1 = columns(x);

	MED = (d~x)*B50;
	IQRANGE = max((d~x)*B75_B25,1e-8);

	uhat = (y - MED)./IQRANGE;
	ubar = meanc(uhat);
	varu = varc(uhat);
	
	vc = zeros(k*t,k*t);
	b = zeros(k,2*columns(beta));
	vc_g = zeros(r*t,r*t);
	g = zeros(r,2*columns(gamma));
    S = (1/n)*(x~z)'*(x~z);

	for (i = 0; i<t; ++i)
	{
    	for (j = i; j<t; ++j)
		{
        if(i != j){
            yhat1 = (d~x)*(beta[][i]);
            yhat2 = (d~x)*(beta[][j]);
			sparsity = ((ones(n,1)./(IQRANGE*sqrt(varu))).*densn((yhat1 - MED - ubar*IQRANGE)./(IQRANGE*sqrt(varu))));
            V = Kmat(sparsity,x,z,A,tau[0][i]);
			K1 = V[0];
			JT1 = V[1];
			J1 = Jmat(sparsity,d~x,x~z,tau[0][i]);
			sparsity = ((ones(n,1)./(IQRANGE*sqrt(varu))).*densn((yhat2 - MED - ubar*IQRANGE)./(IQRANGE*sqrt(varu))));
			V = Kmat(sparsity,x,z,A,tau[0][j]);
			K2 = V[0];
			JT2 = V[1];
            J2 = Jmat(sparsity,d~x,x~z,tau[0][j]);
            G1 = invert(J1*invert(K1)*J1');
            G2 = invert(J2*invert(K2)*J2');
            H = J1*invert(K1)*S*invert(K2)*J2';
            vc[k*i:k*(i+1)-1][k*j:k*(j+1)-1] = 
               (1/n)*(min(tau[0][i],tau[0][j])-(tau[0][i])*(tau[0][j]))*G1*H*G2;
            vc[k*j:k*(j+1)-1][k*i:k*(i+1)-1] = 
               (1/n)*(min(tau[0][i],tau[0][j])-(tau[0][i])*(tau[0][j]))*G2'*H'*G1';
            if(rows(gamma) == 0){
                g = < >;
                vc_g = < >;
			}
            else{
                IJT1 = invert(JT1);
				J = IJT1[k1:columns(JT1)-1][];
				G = J[0:s-1][];
				H = invert(K1)-((invert(JT1[0:k1-1][0:k1-1]) ~ zeros(k1,r)) |
						 (zeros(r,k1) ~ zeros(r,r)));
				C1 = J*(unit(k1+r) - G'*invert(G*H*G')*G*H);
                IJT2 = invert(JT2);
				J = IJT2[k1:columns(JT2)-1][];
				G = J[0:s-1][];
				H = invert(K2)-((invert(JT2[0:k1-1][0:k1-1]) ~ zeros(k1,r)) |
						 (zeros(r,k1) ~ zeros(r,r)));
				C2 = J*(unit(k1+r) - G'*invert(G*H*G')*G*H);
                vc_g[r*i:r*(i+1)-1][r*j:r*(j+1)-1] = 
                     (1/n)*(min(tau[0][i],tau[0][j])-tau[0][i]*tau[0][j])*C1*S*C2';
                vc_g[r*j:r*(j+1)-1][r*i:r*(i+1)-1] = 
                     (1/n)*(min(tau[0][i],tau[0][j])-tau[0][i]*tau[0][j])*C2*S*C1';
            }
		}	
        else{
            yhat = (d~x)*beta[][i];
			sparsity = ((ones(n,1)./(IQRANGE*sqrt(varu))).*densn((yhat - MED - ubar*IQRANGE)./(IQRANGE*sqrt(varu))));
			V = Kmat(sparsity,x,z,A,tau[0][i]);
			K = V[0];
			JT = V[1];
			J = Jmat(sparsity,d~x,x~z,tau[0][i]);
            G = invert(J*invert(K)*J');
            H = J*invert(K)*S*invert(K)*J';
            vc[k*i:k*(i+1)-1][k*i:k*(i+1)-1] = 
               (1/n)*(tau[0][i]-(tau[0][i])^2)*G*H*G;
            b[][2*i] = beta[][i];
            b[][2*i+1] = (sqrt(diagonal(vc[k*i:k*(i+1)-1][k*i:k*(i+1)-1])))';
            if(rows(gamma) == 0){
                g = < >;
                vc_g = < >;
			}
            else{
                IJT = invert(JT);
				J = IJT[k1:columns(JT)-1][];
				G = J[0:s-1][];
				H = invert(K)-((invert(JT[0:k1-1][0:k1-1]) ~ zeros(k1,r)) |
						 (zeros(r,k1) ~ zeros(r,r)));
				C = J*(unit(k1+r) - G'*invert(G*H*G')*G*H);
                vc_g[r*i:r*(i+1)-1][r*i:r*(i+1)-1] =
                     (1/n)*(tau[0][i]-(tau[0][i])^2)*C*S*C';
                g[][2*i] = gamma[][i];
                g[][2*i+1] = (sqrt(diagonal(vc_g[r*i:r*(i+1)-1][r*i:r*(i+1)-1])))';
            }
			if(i == 0){
				J_t = JT;
			}
			else{
				J_t = J_t | JT;
			}
        }      
    	}
	}

	result[0] = b;
	result[1] = vc;
	result[2] = g;
	result[3] = vc_g;
	result[4] = J_t;

	return(result);

}