//*********************************************************************
// RQ Ox  version 1.0 August 1999.  Modified 12/20/2001 CBH
// This is a group of Ox programs to solve unconstrained quantile regression 
// problems using a preprocessing approach described in Portnoy and Koenker
// (1997) and  interior point methods. The Ox translation of the interior point
// algorithm is by Daniel Morillo and Roger Koenker of Splus and ratfor code by R.
// Koenker. The implementation of the parametric programming and inference procedures
// is by Daniel Morillo.
// See Portnoy and Koenker (1997) for a detailed description of the algorithm.
// The program now includes two functions that solve parametric programming
// problems once the regression is computed. This allows the computation of
// coefficients for a range of tau as well as inference on the obtained
// coefficients
//
// The programs in this release comprise the following:
//
//	lp_fnm: Frisch-Newton-Mehrotra algorithm for solving bounded
//		variables linear programs.
//	rq_fnm: Calling routine for using lp_fnm to solve rq problems.
//	sparsity: A simple function to compute Siddiqui estimate of
//		the sparsity function (reciprocal of the density). 
//	RQ: This function implements the algorithm described in
//		Portnoy and Koenker (1997).  It calls rq_fnm after
//		some preprocessing to reduce the effective dimension of
//		the problem.
//	rq: This is the "wrapper function" that the end user should use.
//
//
//	NOTE: The end user should be able to use just the function "rq" for most
//		purposes. All functions and global variables that are used to change
//		varios options are fully documented in the code.
//
// This version of the functions were tested in Version 2.1 of Ox. See
// http://www.nuff.ox.ac.uk/Users/Doornik/ for details on Ox.  The paper
// by Portnoy and Koenker and further details of the Splus/Fortran
// implementation of the algorithm are available from 
// http://www.econ.uiuc.edu/~roger/research/rqn/rqn.html.

//**********************************************************************

#include <oxstd.h>
#include <oxprob.h>

//*****************************************************************************
// GLOBAL VARIABLES:
//
// VERBOSE: Controls output during program operation. Set to zero for no
//	intermediate output. Up to 3 for maximum verbosity.
//*****************************************************************************

decl VERBOSE=0;

//**********************************************************************
// Function: lp_fnm
//
// Description:
// This function is designed to solve the bounded variables linear program:
//
//   min_x {c'x | Ax = b,  0 <= x <= u}
//
// Data:
//
//   A  is a p by n matrix
//   c  is a n vector
//   b  is a p vector
//   u  is a n vector
//
// Control Parameters:
//
//   eps is a scalar termination criterion parameter
//   beta is a scalar step length parameter
//
// VERSION NOTES:
//
//  1) This is a Frisch-Newton-Mehrotra algorithm for solving bounded variables
//    linear programs.  This designation is based on the following attributions
//    Frisch(1956) who apparently first suggested log-barrier methods for LP's
//    Newton(?) who apparently first suggested Newton's method
//    Mehrotra(1992) who apparently first suggested the particular updating 
//    used here for the barrier parameter, mu.
//  2) Many others obviously had a hand in developing the ideas underlying 
//    the algorithm, among whom we might mentioned Kantorovich's
//    analysis of Newton's method, and Fiacco and McCormick's
//    development of barrier methods.  Many details of the 
//    implementation used here are based on the implementation
//    of Lustig, Marsden and Shanno(1994).
//  3) The Ox version is based on a translation of an Splus/ratfor
//    version written by R. Koenker, and described in Portnoy
//    and Koenker (1997).  The Ox translation was carried out by
//    Daniel Morillo and Roger Koenker. 
//  4) Contrary to LMS, we assume that an initial _feasible_ solution
//    is available.  This simplifies somewhat the rhs's.
//  5) The updating of mu reverts to the original Mehrotra proposal.
//  6) Convergence is based on the unnormalized duality gap.
//  7) Equation numbers below refer to LMS (1994).
//  8) All vectors are _row_ vector for annoying Ox reasons of efficiency.
//    except for dy which needs to be a column vector....aaargh!
//
// Syntax:
//
//	Input:
//		A, p by n matrix
//  	c, n row vector
//  	b, p row vector
//  	u, n row vector of upper bounds on the dual x
//  	x, address:
//			in: starting values for dual n row vector
//			out: dual n row vector
//
//	Output:
//		p column vector of primal variables at the solution
//		
//**********************************************************************
lp_fnm(const A,const c,const b,const u,x){
	decl s,w,z,y,r,q,dx,ds,dw,dz,dy,xinv,sinv,xi,dxdz,dsdw;
	decl mu,gap,delta_p,delta_d,g;
	decl rhs,L,D;
	decl it,n;
			
	decl BIG =   1e+20; //a large number
	decl MAX_IT   = 50;	//maximum number of iterations
	decl eps =   1e-05;	//convergence precision
	decl beta = 0.9995;	//step length factor

	n = columns(A);
	
	// generate initial feasible point

	if(VERBOSE>=3){
		print("\n      Starting interior point algorithm\n");
	}

	s = u-x[0];				

	ols2r(c,A,&y);			// Major Choleski's OLS

	r = c - y*A;			// residual from OLS regression
	z =  r .> 0 .? r .: 0;		// z is + part of r, w is - part of r
	w =  z - r;			// r=z-w this guarantees feasibility 

	gap = (c*x[0]'-b*y'+u*w');		// duality gap

	// loop starts	

	it  =  0;
	while( gap > eps && it < MAX_IT ){
		
		++it;
		if(VERBOSE>=3){
			print("        Doing iteration #",it," gap is ",double(gap),"\n");
		}
				
		// solve  (19) for affine step

		q = (1)./(z./x[0]+w./s);
		r = z-w;
		rhs = (q.*r)*A';
		decldl((A.*q)*A',&L,&D);// Choleski decomposition
		dy = solveldl(L,D,rhs');// backsolve for dy, NB. dy is p by 1

		//compute the other increments

		dx = q.*(dy'*A-r);
		ds = -dx;
		dz = (-z.*dx)./x[0] - z;
		dw = (-w.*ds)./s - w;

		//compute step sizes, equation (20)

		delta_p =  min((dx .< 0 .? (x[0]./(-dx)) .: BIG),
				(ds .< 0 .? (s./(-ds)) .: BIG));
		delta_d =  min((dz .< 0 .? (z./(-dz)) .: BIG),
				(dw .< 0 .? (w./(-dw)) .: BIG));

		delta_p = min(beta*delta_p,1);
		delta_d = min(beta*delta_d,1);

		//If full affine step is feasible take it, otherwise modify it.

		if(min(delta_p,delta_d)<1){

			// update mu

			mu = x[0]*z' + s*w';
			g = (x[0]+delta_p*dx)*(z+delta_d*dz)' + 
				(s+delta_p*ds)*(w+delta_d*dw)';
			mu = ((g/mu)^3)*(mu/(2*n));

			// now solve (24) for modified step

			dxdz = dx.*dz;
			dsdw = ds.*dw;
			xinv = (1)./x[0];
			sinv = (1)./s;

			xi = mu*(xinv-sinv);
			rhs = rhs+(q.*((dxdz-dsdw)-xi))*A';
			dy = solveldl(L,D,rhs');//backsolve with new rhs

			dx = q.*(dy'*A+xi-r-(dxdz-dsdw));
			ds = -dx;
			dz = mu*xinv-z-xinv.*z.*dx-dxdz;
			dw = mu*sinv-w-sinv.*w.*ds-dsdw;

			// now compute new step lengths, again as in (24)

			delta_p =  min((dx .< 0 .? (x[0]./(-dx)) .: BIG),
					(ds .< 0 .? (s./(-ds)) .: BIG));
			delta_d =  min((dz .< 0 .? (z./(-dz)) .: BIG),
					(dw .< 0 .? (w./(-dw)) .: BIG));

			delta_p = min(beta*delta_p,1);
			delta_d = min(beta*delta_d,1);
		}

		// now update all variables

		x[0] = x[0]+delta_p*dx;
		s = s+delta_p*ds;
		y = y+delta_d*dy';
		w = w+delta_d*dw;
		z = z+delta_d*dz;

		gap = (c*x[0]'-b*y'+u*w');
	}
	if(VERBOSE>=3){
		print("\n");
	}
	return(y');
}

//*********************************************************************
// Function: rq_fnm
//
// Description:
// Solves unconstrained quantile regression problems using a primal-dual
// interior point method.  Translation by Daniel Morillo and Roger Koenker of
// splus and ratfor code by R. Koenker.
//
// Syntax:
//
//	Input:
//
//		y, n column vector of n dependent observations
//		X, n by p matrix of n observations of p variables. No constant is included
//   		tau, scalar between 0 and 1 determining the quantile to estimate
//	 	a, address:
//			in: nothing
//			out: dual n row vector at the solution
//
//
// 	Output:
//   	p column vector of parameter estimates
//
//*******************************************************************
rq_fnm(const X, const y, const tau,const a){
	decl n,p,b,u;

	n = rows(X);		// sample size
	p = columns(X);		// dimensionality of beta
	u = ones(1,n);		// fixed upper bounds for dual variables
	a[0] = (1-tau)*u;	// starting values for dual variables

	b = lp_fnm(X',-y',a[0]*X,u,a); //a needs to be address
	a[0]=a[0]';
	
	return(-b);
}

//*******************************************************************
// Function: sparsity
//
// Description:
// Siddiqui estimate of the sparsity function
// A simple difference quotient estimate of the reciprocal of the
// density at a particular quantile.  This function is used to
// estimate a nuisance parameter for quantile regression inference.
//
// Syntax:
//
//	Input:
//  	r, column vector of residuals 
//  	tau, scalar determining the quantile of interest
//
//	Output:
//  	scalar, the sparsity estimate
//
//*******************************************************************
sparsity(const r, const tau){
        decl  h,qhat;
        decl n;
 
        n=rows(r);
        h=(0.05+3.65*tau-3.65*(tau^2))*(n^(2/3))/n; // Hall-Sheather bandwidth 
        qhat=quantilec(r,tau + h .* <-1 , 1>);
        return((qhat[1][0]-qhat[0][0])/(2*h)); 
}

//*********************************************************************
// Function: RQ
//
// Description:
// Solves unconstrained quantile regression problems using a primal-dual
// interior point method and the preprocessing approach described in
// Portnoy and Koenker(1997).  The Ox code is a translation by Daniel 
// Morillo and Roger Koenker of Splus and ratfor code by R. Koenker.
//
// Syntax:
//
// Input:
//   y, n column vector of independent observations
//   X, n by p matrix of n dependent observations in p variables. No constant is included
//   tau, scalar between 0 and 1 determining the quantile to estimate
//
// Output:
//   p column vector of parameter estimates
//
//*******************************************************************

RQ(X,y, const tau){
	decl n,p,m,it;
	decl not_optimal,bad_fixups;
	decl XX,yy,b,r,hi,d;
	decl su,sl,in,su_bad,sl_bad;
	decl a,S,Sinv,cs,qb,discr,b0;
	decl sortvec;

	decl MAX_BAD_FIXUPS = 5;
	decl MAX_IT = 5;

	n = rows(X);            // sample size
    p = columns(X);         // dimensionality of beta
	m = floor(max(n/20,n^(2/3)));	// preliminary sample size
	not_optimal = TRUE;
	it = 0;

	if(VERBOSE>=1){
		print("\nStarting Globbing procedure\n");
	}
	
	sortvec=rann(n,1);
	X=sortbyc(sortvec~X,0)[][1:p];
	y=sortbyc(sortvec~y,0)[][1];

	while(not_optimal && it < MAX_IT){
		++it;
		m  = min(2*m,n-1);

		if(VERBOSE>=2){
			print("  Now doing iteration #",it," with ",m," observations\n");
		}
		XX = X[0:m][];
		yy = y[0:m][];
		b  = rq_fnm(XX,yy,tau,&a);	//preliminary estimation
		r  = y - X*b;
		cs = 2*sparsity(r,tau)*sqrt(tau*(1-tau));
		Sinv=XX'*XX;
		S=invertsym(Sinv);
		hi = sqrt(sumsqrr(X*choleski(S)));
		d  = cs*hi;
		su = r .>  d; 		//above the band
		sl = r .< -d;		//below the band
		bad_fixups = 0;
		while(not_optimal && (bad_fixups < MAX_BAD_FIXUPS)){
			in = !su .&& !sl;	//inside the band
			XX = selectifr(X,in);
			yy = selectifr(y,in);
		//make the globs
			if(any(su)){
				XX = XX | sumc(selectifr(X,su));
				yy = yy | sumc(selectifr(y,su));
				}
			if(any(sl)){
				XX = XX | sumc(selectifr(X,sl));
				yy = yy | sumc(selectifr(y,sl));
			}
		//reestimate with globbed sample
			b0=b;
			if(VERBOSE>=2){
				print("    Re-estimating with globbed sample of ",rows(XX)," observations\n");
			}
			if(rank(XX)==p){
				b = rq_fnm(XX,yy,tau,&a);
			}
			else{
				b = bad_fixups*(ones(p,1));
			}
			r = y - X*b;
		//check the globs
			qb=b-b0;
			discr=cs^2-qb'*Sinv*qb;
			if((bad_fixups==0 && discr<=0) || bad_fixups>0){
				su_bad = (r .<= 0) .&& su;
				sl_bad = (r .>= 0) .&& sl;
			
				if(any(su_bad ~ sl_bad)){
					su = su .&& !su_bad;
					sl = sl .&& !sl_bad;
					++bad_fixups;
					if(VERBOSE>=2){
						print("    Bad observations found!, trying again\n");
					}
				}
				else{ 
					not_optimal = FALSE;
				}
			}
			else{
				not_optimal=FALSE;
			}
		}	
	}
	return(b);
}

//******************************************************************************
// Function: rq
//
// Description:
// This is just a wrapper function.
//
// Syntax:
//
//	Input:
//		y, n column vector of observations of the dependent variable
//		X, x by p matrix of n observations of p independent variables. A constant is
//         automatically included
//		tau, scalar in (0,1) corresponding ot the quantile to estimate.
//		GLOB, scalar. 1 for preprocessing, 0 for no preprocessing.
//
//	Output:
//      Function returns only parameter estimates.  Standard errors/
//      confidence intervals are not computed.
//
//********************************************************************************
rq(X,y,tau,GLOB){
	decl result,n,tmp1;

	n=rows(y);

	if(VERBOSE>=1){
		print("\nStarting estimation procedure\n");
		print("Options:\n");
		print("  Computing single tau=",tau,"\n");
		if(GLOB==1){
			print("  Globbing procedures will be used\n");
		}
		print("  Verbose mode is ",VERBOSE,"\n");
	}
	
	if(rows(X) == 0){
		result=quantilec(y,tau);
	}
	else if(GLOB==1){
		result=RQ(X~ones(n,1),y,tau);
	}
	else{
		result=rq_fnm(X~ones(n,1),y,tau,&tmp1);
	}
	
	return(result);
}
