#include <oxstd.h>
#include <oxfloat.h>
#include <oxprob.h>

//******************************************************************************
// Function: kcov_iqr
//
// [b,vc] = kcov_iqr(beta,tau,y,d,x,z)
// Calculates the heteroskedasticity consistent variance-covariance
//   matrix for inverse quantile regression estimates using kernel methods
//   as introduced in Powell (1984) and Buchinsky (1995). Note that this
//   implementation is for just-id models only.
// Output --
//   b - coefficient vector and standard errors
//       If betas for more than one quantile are passed, the odd rows
//       are the coefficient estimates and the even rows are the standard
//       errors.
//   vc - the complete variance-covariance matrix
//       If more than one quantile is passed this is the variance-covariance
//       matrix of the joint distribution for all of the quantiles passed.
//
// Input --
//   beta - coefficient estimates from IQR.  Estimates from more than one
//          quantile may be passed.  Each column should contain the estimates
//          for one quantile.
//   tau - row vector containing the quantile index of each column of beta/gamma
//   y - observations on dependent variable 
//   d - observations on RHS endogenous variable 
//   x - observations on covariates 
//   NOTE:  x should not include a constant term.
//   z - observations on excluded instruments


kcov_iqr(beta,tau,y,d,x,z){

	decl b, vc, n, S, e, J, result, h, k;

	result = new array[3];

	n = rows(y);	 
	x = x~ones(n,1);
	k = columns(d~x);

    vc = zeros(k,k);
	b = zeros(k,2);
    S = (1/n)*(x~z)'*(x~z);

    e = y - (d~x)*beta;

   	h = 1.364*((2*sqrt(M_PI))^(-1/5))*sqrt(varc(e))*(n^(-1/5));
   	J = (1/(n*h))*(densn(e/h).*(d~x))'*(x~z);

	vc = (1/n)*(tau-tau^2)*invert(J')*S*invert(J);
    b[][0] = beta;
    b[][1] = (sqrt(diagonal(vc)))';

	J = invert(J);
	
	result[0] = b;
	result[1] = vc;
	result[2] = J;

	return(result);

}