#include <oxstd.h>
#include <oxfloat.h>
#include <oxprob.h>

//******************************************************************************
// Function: kcov_qr
//
// [b,vc] = kcov_qr(beta,tau,y,x)
// Calculates the heteroskedasticity consistent variance-covariance
//   matrix for inverse quantile regression estimates using kernel methods
//   as introduced in Powell (1984) and Buchinsky (1995).
// Output --
//   b - coefficient vector and standard errors
//       If betas for more than one quantile are passed, the odd rows
//       are the coefficient estimates and the even rows are the standard
//       errors.
//   vc - the complete variance-covariance matrix
//       If more than one quantile is passed this is the variance-covariance
//       matrix of the joint distribution for all of the quantiles passed.
//
// Input --
//   beta - coefficient estimates from IQR.  Estimates from more than one
//          quantile may be passed.  Each column should contain the estimates
//          for one quantile.
//   tau - row vector containing the quantile index of each column of beta/gamma
//   y - observations on dependent variable (n X 1)
//   x - observations on covariates (n X k)
//   NOTE:  x should not include a constant term.


kcov_qr(beta,tau,y,x){

	decl b, vc, n, S, e, J, k, result, h;

	result = new array[3];

	n = rows(y);	 
	x = x~ones(n,1);
	k = rows(beta);

    vc = zeros(k,k);
	b = zeros(k,2);
    S = (1/n)*(x)'*(x);

    e = y - x*beta;

	h = 1.364*((2*sqrt(M_PI))^(-1/5))*sqrt(varc(e))*(n^(-1/5));
   	J = ((1/(n*h))*(densn(e/h).*x)'*x)^(-1);

    vc = (1/n)*(tau-tau^2)*J*S*J;
    b[][0] = beta;
    b[][1] = (sqrt(diagonal(vc)))';

	result[0] = b;
	result[1] = vc;
	result[2] = J;

	return(result);

}